using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.Linq;

namespace ksp_osm
{
    struct SingleOrArray<T>
    {
        private object val;

        public SingleOrArray(T[] arr)
        {
            if (arr.Length == 1)
                this.val = arr[0];
            else
                this.val = arr;
        }

        public T[] AsArray => val is T valT ? new T[] { valT } : (T[])val;
        public int Length => val is T ? 1 : ((T[])val).Length;
    }
    public class Polygon
    {
        private readonly SingleOrArray<Point[]> ways;
        /// Obdélník obsahující celý polygon. Hodí se na rychlou kontrolu, jestli má cenu se vůbec bodem zabyvat.
        public readonly Rectangle Boundary;
        public long Id { get; set; }

        public Polygon(Point[][] ways)
        {
            this.ways = new SingleOrArray<Point[]>(ways);
            var allpoints = ways.SelectMany(a => a);
            this.Boundary = Rectangle.FromLTRB(
                allpoints.Min(a => a.X),
                allpoints.Min(a => a.Y),
                allpoints.Max(a => a.X),
                allpoints.Max(a => a.Y));
        }

        public bool Contains(Point loc)
        {
            // rychlá kontrola
            if (!this.Boundary.Contains(loc))
                return false;

            // načrtneme horizontální čáru od bodu `loc` a podíváme se kolikrát protíná stranu mnohoúhelníku
            var horizontalLine = loc.Y;
            var crossCount = 0;
            foreach (var w in this.ways.AsArray)
            {
                for (int i = 1; i < w.Length; i++)
                {
                    // zkontrolujeme stranu z w[i - 1] do w[i]

                    // pokud jsou vrcholy strany na opačných stranách od naší horizontální čáry
                    if ((w[i - 1].Y < horizontalLine) != (w[i].Y < horizontalLine))
                    {
                        var s = (from: w[i - 1], to: w[i]);
                        var relativeCrossPoint =
                            (double)(s.from.Y - horizontalLine) / (double)(s.from.Y - s.to.Y);
                        Debug.Assert(relativeCrossPoint >= 0 && relativeCrossPoint <= 1);
                        var crossPointX =
                            s.from.X + ((s.to.X - s.from.X) * relativeCrossPoint);

                        if (crossPointX >= loc.X)
                            crossCount += 1;
                    }
                }
            }

            // když jsme prošli stranou mnohoúhelníku lišekrát, jsme uvnitř
            return crossCount % 2 == 1;
        }

        public bool IntersectsWith(Rectangle rect)
        {
            // we don't have anything in common
            if (!this.Boundary.IntersectsWith(rect))
                return false;

            // either, the polygon contains one corner, or the rectangle contains some corner of the polygon

            foreach (var w in this.ways.AsArray)
                foreach (var p in w)
                    if (rect.Contains(p))
                        return true;

            var p1 = new Point(rect.Left, rect.Top);
            var p2 = new Point(rect.Left, rect.Bottom);
            var p3 = new Point(rect.Right, rect.Top);
            var p4 = new Point(rect.Right, rect.Bottom);

            return Contains(p1) || Contains(p2) || Contains(p3) || Contains(p4);
        }

        public static Polygon FromRelation(OsmMap map, long relId)
        {
            var ways = new List<Way>();
            void loadRel(long id)
            {
                foreach (var m in map.Relations[id].Members)
                {
                    if (m.Type == OsmType.Way)
                        ways.Add(map.Ways[m.Id]);
                    else if (m.Type == OsmType.Relation)
                        loadRel(m.Id);
                }
            }
            loadRel(relId);
            return new Polygon(ways.Select(w => w.NodeRefs.Select(n => map.Nodes[n].Location).ToArray()).ToArray()) { Id = relId };
        }

        public static Polygon FromWay(OsmMap map, long wayId)
        {
            var way = map.Ways[wayId];
            Debug.Assert(way.NodeRefs.Length >= 3);
            Debug.Assert(way.NodeRefs.First() == way.NodeRefs.Last());

            var points = way.NodeRefs.Select(n => map.Nodes[n].Location).ToArray();
            return new Polygon(new [] { points }) { Id = wayId };
        }
    }
}
